@@ -17,17 +17,7 @@
 
    Written by Torbjorn Granlund, Sweden (tege@sics.se). */
 
-#ifdef HAVE_CONFIG_H
-#if defined (CONFIG_BROKETS)
-/* We use <config.h> instead of "config.h" so that a compilation
-   using -I. -I$srcdir will use ./config.h rather than $srcdir/config.h
-   (which it would do because it found this file in $srcdir).  */
 #include <config.h>
-#else
-#include "config.h"
-#endif
-#endif
-
 #include <stdio.h>
 
 #include "cp.h"
@@ -69,8 +59,8 @@ Usage: %s [OPTION]... SOURCE DEST\n\
   -x, --one-file-system        stay on this file system\n\
   -P, --parents                append source path to DIRECTORY\n\
   -R, --recursive              copy directories recursively\n\
-  -S, --suffix SUFFIX          override the usual backup suffix\n\
-  -V, --version-control WORD   override the usual version control\n\
+  -S, --suffix=SUFFIX          override the usual backup suffix\n\
+  -V, --version-control=WORD   override the usual version control\n\
       --help                   display this help and exit\n\
       --version                output version information and exit\n\
 \n\
