@@ -20,17 +20,21 @@
 #ifdef _AIX
  #pragma alloca
 #endif
+
+#include <config.h>
 #include <stdio.h>
 #include <getopt.h>
 #include "cp.h"
 #include "backupfile.h"
 #include "version.h"
+#include "safe-stat.h"
+#include "safe-lstat.h"
 
 #ifndef _POSIX_VERSION
 uid_t geteuid ();
 #endif
 
-/* Used by do_copy, make_path, and re_protect
+/* Used by do_copy, make_path_private, and re_protect
    to keep a list of leading directories whose protections
    need to be fixed after copying. */
 struct dir_attr
@@ -41,13 +45,15 @@ struct dir_attr
 };
 
 char *dirname ();
+char *xstrdup ();
 enum backup_type get_version ();
 int eaccess_stat ();
+int full_write ();
 
 static int do_copy ();
 static int copy ();
 static int copy_dir ();
-static int make_path ();
+static int make_path_private ();
 static int copy_reg ();
 static int re_protect ();
 
@@ -64,8 +70,8 @@ char *program_name;
    whether dereferencing of symlinks is done.  */
 static int (*xstat) ();
 
-/* If nonzero, copy all files except directories and, if not dereferencing
-   them, symbolic links, as if they were regular files. */
+/* If nonzero, copy all files except (directories and, if not dereferencing
+   them, symbolic links,) as if they were regular files. */
 static int flag_copy_as_regular = 1;
 
 /* If nonzero, dereference symbolic links (copy the files they point to). */
@@ -119,10 +125,10 @@ static int umask_kill;
 static uid_t myeuid;
 
 /* If non-zero, display usage information and exit.  */
-static int flag_help;
+static int show_help;
 
-/* If non-zero, print the version on standard error.  */
-static int flag_version;
+/* If non-zero, print the version on standard output and exit.  */
+static int show_version;
 
 static struct option const long_opts[] =
 {
@@ -133,6 +139,7 @@ static struct option const long_opts[] =
   {"link", no_argument, NULL, 'l'},
   {"no-dereference", no_argument, &flag_dereference, 0},
   {"one-file-system", no_argument, &flag_one_file_system, 1},
+  {"parents", no_argument, &flag_path, 1},
   {"path", no_argument, &flag_path, 1},
   {"preserve", no_argument, &flag_preserve, 1},
   {"recursive", no_argument, NULL, 'R'},
@@ -141,8 +148,8 @@ static struct option const long_opts[] =
   {"update", no_argument, &flag_update, 1},
   {"verbose", no_argument, &flag_verbose, 1},
   {"version-control", required_argument, NULL, 'V'},
-  {"help", no_argument, &flag_help, 1},
-  {"version", no_argument, &flag_version, 1},
+  {"help", no_argument, &show_help, 1},
+  {"version", no_argument, &show_version, 1},
   {NULL, 0, NULL, 0}
 };
 
@@ -254,21 +261,21 @@ main (argc, argv)
 	  break;
 
 	default:
-	  usage ((char *) 0);
+	  usage (2, (char *) 0);
 	}
     }
 
-  if (flag_version)
+  if (show_version)
     {
-      fprintf (stderr, "%s\n", version_string);
+      printf ("%s\n", version_string);
       exit (0);
     }
 
-  if (flag_help)
-    usage (NULL);
+  if (show_help)
+    usage (0, NULL);
 
   if (flag_hard_link && flag_symbolic_link)
-    usage ("cannot make both hard and symbolic links");
+    usage (2, "cannot make both hard and symbolic links");
 
   if (make_backups)
     backup_type = get_version (version);
@@ -280,9 +287,9 @@ main (argc, argv)
      of `stat' to call.  */
 
   if (flag_dereference)
-    xstat = stat;
+    xstat = safe_stat;
   else
-    xstat = lstat;
+    xstat = safe_lstat;
 
   /* Allocate space for remembering copied and created files.  */
 
@@ -307,13 +314,13 @@ do_copy (argc, argv)
   int ret = 0;
 
   if (optind >= argc)
-    usage ("missing file arguments");
+    usage (2, "missing file arguments");
   if (optind >= argc - 1)
-    usage ("missing file argument");
+    usage (2, "missing file argument");
 
   dest = argv[argc - 1];
 
-  if (lstat (dest, &sb))
+  if (SAFE_LSTAT (dest, &sb))
     {
       if (errno != ENOENT)
 	{
@@ -330,7 +337,7 @@ do_copy (argc, argv)
       /* If `dest' is not a symlink to a nonexistent file, use
 	 the results of stat instead of lstat, so we can copy files
 	 into symlinks to directories. */
-      if (stat (dest, &sbx) == 0)
+      if (SAFE_STAT (dest, &sbx) == 0)
 	sb = sbx;
     }
 
@@ -358,10 +365,10 @@ do_copy (argc, argv)
 	      dst_path = xmalloc (strlen (dest) + strlen (arg) + 2);
 	      stpcpy (stpcpy (stpcpy (dst_path, dest), "/"), arg);
 
-	      /* For --path, we have to make sure that the directory
+	      /* For --parents, we have to make sure that the directory
 	         dirname (dst_path) exists.  We may have to create a few
 	         leading directories. */
-	      parent_exists = !make_path (dst_path,
+	      parent_exists = !make_path_private (dst_path,
 					  strlen (dest) + 1, 0700,
 					  flag_verbose ? "%s -> %s\n" :
 					  (char *) NULL,
@@ -374,7 +381,7 @@ do_copy (argc, argv)
 	      ap = basename (arg);
 	      /* For `cp -R source/.. dest', don't copy into `dest/..'. */
 	      if (!strcmp (ap, ".."))
-		dst_path = dest;
+		dst_path = xstrdup (dest);
 	      else
 		{
 		  dst_path = xmalloc (strlen (dest) + strlen (ap) + 2);
@@ -384,14 +391,14 @@ do_copy (argc, argv)
 
 	  if (!parent_exists)
 	    {
-	      /* make_path failed, so we shouldn't even attempt the copy. */
+	      /* make_path_private failed, so we shouldn't even attempt the copy. */
 	      ret = 1;
   	    }
 	  else
 	    {
 	      ret |= copy (arg, dst_path, new_dst, 0, (struct dir_list *) 0);
 	      forget_all ();
-  
+
 	      if (flag_path)
 		{
 		  ret |= re_protect (dst_path, strlen (dest) + 1,
@@ -399,6 +406,7 @@ do_copy (argc, argv)
 		}
 	    }
 
+	  free (dst_path);
 	  ++optind;
 	  if (optind == argc - 1)
 	    break;
@@ -407,12 +415,45 @@ do_copy (argc, argv)
     }
   else if (argc - optind == 2)
     {
+      char *new_dest;
+      char *source;
+      struct stat source_stats;
+
       if (flag_path)
-	usage ("when preserving paths, last argument must be a directory");
-      return copy (argv[optind], dest, new_dst, 0, (struct dir_list *) 0);
+	usage (2, "when preserving paths, last argument must be a directory");
+
+      source = argv[optind];
+
+      /* When the destination is specified with a trailing slash and the
+	 source exists but is not a directory, convert the user's command
+	 `cp source dest/' to `cp source dest/basename(source)'.  */
+
+      if (dest[strlen (dest) - 1] == '/'
+	  && SAFE_LSTAT (source, &source_stats) == 0
+	  && !S_ISDIR (source_stats.st_mode))
+	{
+	  char *source_base;
+	  char *tmp_source;
+
+	  tmp_source = (char *) alloca (strlen (source) + 1);
+	  strcpy (tmp_source, source);
+	  strip_trailing_slashes (tmp_source);
+	  source_base = basename (tmp_source);
+
+	  new_dest = (char *) alloca (strlen (dest) + 1 +
+				      strlen (source_base) + 1);
+	  stpcpy (stpcpy (stpcpy (new_dest, dest), "/"), source_base);
+	}
+      else
+	{
+	  new_dest = dest;
+	}
+
+      return copy (source, new_dest, new_dst, 0, (struct dir_list *) 0);
     }
   else
-    usage ("when copying multiple files, last argument must be a directory");
+    usage (2,
+	   "when copying multiple files, last argument must be a directory");
 }
 
 /* Copy the file SRC_PATH to the file DST_PATH.  The files may be of
@@ -514,10 +555,11 @@ copy (src_path, dst_path, new_dst, device, ancestors)
 	    {
 	      if (flag_interactive)
 		{
-		  if (eaccess_stat (&dst_sb, W_OK) != 0)
+		  if (eaccess_stat (&dst_sb, W_OK, dst_path) != 0)
 		    fprintf (stderr,
 			     "%s: overwrite `%s', overriding mode %04o? ",
-			     program_name, dst_path, dst_sb.st_mode & 07777);
+			     program_name, dst_path,
+			     (unsigned int) (dst_sb.st_mode & 07777));
 		  else
 		    fprintf (stderr, "%s: overwrite `%s'? ",
 			     program_name, dst_path);
@@ -711,17 +753,13 @@ copy (src_path, dst_path, new_dst, device, ancestors)
     }
   else
 #ifdef S_ISLNK
-#ifdef _AIX
-#define LINK_BUF PATH_MAX
-#else
-#define LINK_BUF src_sb.st_size
-#endif
   if (S_ISLNK (src_type))
     {
-      char *link_val = (char *) alloca (LINK_BUF + 1);
+      char *link_val;
       int link_size;
 
-      link_size = readlink (src_path, link_val, LINK_BUF);
+      link_val = (char *) alloca (PATH_MAX + 2);
+      link_size = readlink (src_path, link_val, PATH_MAX + 1);
       if (link_size < 0)
 	{
 	  error (0, errno, "cannot read symbolic link `%s'", src_path);
@@ -762,7 +800,9 @@ copy (src_path, dst_path, new_dst, device, ancestors)
 
       /* If non-root uses -p, it's ok if we can't preserve ownership.
 	 But root probably wants to know, e.g. if NFS disallows it.  */
-      if (chown (dst_path, src_sb.st_uid, src_sb.st_gid)
+      if (chown (dst_path,
+		 (myeuid == 0 ? src_sb.st_uid : myeuid),
+		 src_sb.st_gid)
 	  && (errno != EPERM || myeuid == 0))
 	{
 	  error (0, errno, "%s", dst_path);
@@ -801,7 +841,7 @@ un_backup:
 }
 
 /* Ensure that the parent directory of CONST_DIRPATH exists, for
-   the --path option.
+   the --parents option.
 
    SRC_OFFSET is the index in CONST_DIRPATH (which is a destination
    path) of the beginning of the source directory name.
@@ -819,8 +859,8 @@ un_backup:
    permissions when done, otherwise 1. */
 
 static int
-make_path (const_dirpath, src_offset, mode, verbose_fmt_string,
-	      attr_list, new_dst)
+make_path_private (const_dirpath, src_offset, mode, verbose_fmt_string,
+		   attr_list, new_dst)
      char *const_dirpath;
      int src_offset;
      int mode;
@@ -839,7 +879,7 @@ make_path (const_dirpath, src_offset, mode, verbose_fmt_string,
 
   src = dirpath + src_offset;
 
-  tmp_dst_dirname = dirname (dirpath); 
+  tmp_dst_dirname = dirname (dirpath);
   dst_dirname = (char *) alloca (strlen (tmp_dst_dirname) + 1);
   strcpy (dst_dirname, tmp_dst_dirname);
   free (tmp_dst_dirname);
@@ -870,8 +910,9 @@ make_path (const_dirpath, src_offset, mode, verbose_fmt_string,
 	    {
 	      /* This element of the path does not exist.  We must set
 		 *new_dst and new->is_new_dir inside this loop because,
-		 for example, in the command `cp --path ../a/../b/c e_dir',
-		 make_path creates only e_dir/../a if ./b already exists. */
+		 for example, in the command `cp --parents ../a/../b/c e_dir',
+		 make_path_private creates only e_dir/../a if ./b already
+		 exists. */
 	      *new_dst = 1;
 	      new->is_new_dir = 1;
 	      if (mkdir (dirpath, mode))
@@ -911,11 +952,6 @@ make_path (const_dirpath, src_offset, mode, verbose_fmt_string,
       error (0, 0, "`%s' exists but is not a directory", dst_dirname);
       return 1;
     }
-  else if (chmod (dst_dirname, mode))
-    {
-      error (0, errno, "%s", dst_dirname);
-      return 1;
-    }
   else
     {
       *new_dst = 0;
@@ -924,7 +960,7 @@ make_path (const_dirpath, src_offset, mode, verbose_fmt_string,
 }
 
 /* Ensure that the parent directories of CONST_DST_PATH have the
-   correct protections, for the --path option.  This is done
+   correct protections, for the --parents option.  This is done
    after all copying has been completed, to allow permissions
    that don't include user write/execute.
 
@@ -934,7 +970,7 @@ make_path (const_dirpath, src_offset, mode, verbose_fmt_string,
    ATTR_LIST is a null-terminated linked list of structures that
    indicates the end of the filename of each intermediate directory
    in CONST_DST_PATH that may need to have its attributes changed.
-   The command `cp --path --preserve a/b/c d/e_dir' changes the
+   The command `cp --parents --preserve a/b/c d/e_dir' changes the
    attributes of the directories d/e_dir/a and d/e_dir/a/b to match
    the corresponding source directories regardless of whether they
    existed before the `cp' command was given.
@@ -955,7 +991,7 @@ re_protect (const_dst_path, src_offset, attr_list)
 
   dst_path = (char *) alloca (strlen (const_dst_path) + 1);
   strcpy (dst_path, const_dst_path);
-  src_path = dst_path + src_offset; 
+  src_path = dst_path + src_offset;
 
   for (p = attr_list; p; p = p->next)
     {
@@ -1084,7 +1120,6 @@ copy_reg (src_path, dst_path)
   int dest_desc;
   int source_desc;
   int n_read;
-  int n_written;
   struct stat sb;
   char *cp;
   int *ip;
@@ -1137,8 +1172,7 @@ copy_reg (src_path, dst_path)
       /* If the file has fewer blocks than would normally
 	 be needed for a file of its size, then
 	 at least one of the blocks in the file is a hole. */
-      if (S_ISREG (sb.st_mode) &&
-	  sb.st_size - (sb.st_blocks * DEV_BSIZE) >= DEV_BSIZE)
+      if (S_ISREG (sb.st_mode) && sb.st_size > sb.st_blocks * DEV_BSIZE)
 	make_holes = 1;
     }
 #endif
@@ -1152,6 +1186,10 @@ copy_reg (src_path, dst_path)
       n_read = read (source_desc, buf, buf_size);
       if (n_read < 0)
 	{
+#ifdef EINTR
+	  if (errno == EINTR)
+	    continue;
+#endif
 	  error (0, errno, "%s", src_path);
 	  return_val = -1;
 	  goto ret;
@@ -1198,8 +1236,7 @@ copy_reg (src_path, dst_path)
 	}
       if (ip == 0)
 	{
-	  n_written = write (dest_desc, buf, n_read);
-	  if (n_written < n_read)
+	  if (full_write (dest_desc, buf, n_read) < 0)
 	    {
 	      error (0, errno, "%s", dst_path);
 	      return_val = -1;
@@ -1217,12 +1254,12 @@ copy_reg (src_path, dst_path)
     {
 #ifdef HAVE_FTRUNCATE
       /* Write a null character and truncate it again.  */
-      if (write (dest_desc, "", 1) != 1
+      if (full_write (dest_desc, "", 1) < 0
 	  || ftruncate (dest_desc, n_read_total) < 0)
 #else
       /* Seek backwards one character and write a null.  */
       if (lseek (dest_desc, (off_t) -1, SEEK_CUR) < 0L
-	  || write (dest_desc, "", 1) != 1)
+	  || full_write (dest_desc, "", 1) < 0)
 #endif
 	{
 	  error (0, errno, "%s", dst_path);
