@@ -35,6 +35,8 @@
 #ifdef _AIX
  #pragma alloca
 #endif
+
+#include <config.h>
 #include <sys/types.h>
 #if !defined(_POSIX_SOURCE) || defined(_AIX)
 #include <sys/ioctl.h>
@@ -43,28 +45,50 @@
 #include <grp.h>
 #include <pwd.h>
 #include <getopt.h>
+#if HAVE_LIMITS_H
+/* limits.h must come before system.h because limits.h on some systems
+   undefs PATH_MAX, whereas system.h includes pathmax.h which sets
+   PATH_MAX.  */
+#include <limits.h>
+#endif
 #include "system.h"
 #include <fnmatch.h>
-#include "version.h"
 
-#ifndef S_IEXEC
-#define S_IEXEC S_IXUSR
+#include "obstack.h"
+#include "ls.h"
+#include "version.h"
+#include "safe-stat.h"
+#include "safe-lstat.h"
+
+#define obstack_chunk_alloc xmalloc
+#define obstack_chunk_free free
+
+#ifndef INT_MAX
+#define INT_MAX 2147483647
 #endif
 
 /* Return an int indicating the result of comparing two longs. */
-#ifdef INT_16_BITS
+#if (INT_MAX <= 65535)
 #define longdiff(a, b) ((a) < (b) ? -1 : (a) > (b) ? 1 : 0)
 #else
 #define longdiff(a, b) ((a) - (b))
 #endif
 
+/* The maximum number of digits required to print an inode number
+   in an unsigned format.  */
+#ifndef INODE_DIGITS
+#define INODE_DIGITS 7
+#endif
+
 #ifndef STDC_HEADERS
 char *ctime ();
 time_t time ();
+void free ();
 #endif
 
 void mode_string ();
 
+char *stpcpy ();
 char *xstrdup ();
 char *getgroup ();
 char *getuser ();
@@ -123,7 +147,7 @@ enum filetype
   normal			/* All others. */
 };
 
-struct file
+struct fileinfo
 {
   /* The file name. */
   char *name;
@@ -142,13 +166,13 @@ struct file
 
 /* The table of files in the current directory:
 
-   `files' points to a vector of `struct file', one per file.
+   `files' points to a vector of `struct fileinfo', one per file.
    `nfiles' is the number of elements space has been allocated for.
    `files_index' is the number actually in use.  */
 
 /* Address of block containing the files that are described.  */
 
-static struct file *files;
+static struct fileinfo *files;
 
 /* Length of block that `files' points to, measured in files.  */
 
@@ -253,10 +277,17 @@ static int numeric_users;
 static int print_block_size;
 
 /* Nonzero means show file sizes in kilobytes instead of blocks
-   (the size of which is system-dependant).  -k */
+   (the size of which is system-dependent).  -k */
 
 static int kilobyte_blocks;
 
+/* Precede each line of long output (per file) with a string like `m,n:'
+   where M is the number of characters after the `:' and before the
+   filename and N is the length of the filename.  Using this format,
+   Emacs' dired mode starts up twice as fast, and can handle all
+   strange characters in file names.  */
+static int dired;
+
 /* none means don't mention the type of files.
    all means mention the types of all files.
    not_programs means do so except for executables.
@@ -357,16 +388,17 @@ static int format_needs_stat;
 static int exit_status;
 
 /* If non-zero, display usage information and exit.  */
-static int flag_help;
+static int show_help;
 
-/* If non-zero, print the version on standard error.  */
-static int flag_version;
+/* If non-zero, print the version on standard output and exit.  */
+static int show_version;
 
 static struct option const long_options[] =
 {
   {"all", no_argument, 0, 'a'},
   {"escape", no_argument, 0, 'b'},
   {"directory", no_argument, 0, 'd'},
+  {"dired", no_argument, 0, 'D'},
   {"full-time", no_argument, &full_time, 1},
   {"inode", no_argument, 0, 'i'},
   {"kilobytes", no_argument, 0, 'k'},
@@ -389,8 +421,8 @@ static struct option const long_options[] =
   {"sort", required_argument, 0, 10},
   {"tabsize", required_argument, 0, 'T'},
   {"time", required_argument, 0, 11},
-  {"help", no_argument, &flag_help, 1},
-  {"version", no_argument, &flag_version, 1},
+  {"help", no_argument, &show_help, 1},
+  {"version", no_argument, &show_version, 1},
   {0, 0, 0, 0}
 };
 
@@ -421,12 +453,74 @@ static char const* const time_args[] =
   "atime", "access", "use", "ctime", "status", 0
 };
 
+/* This zero-based index is used solely with the --dired option.
+   When that option is in effect, this counter is incremented for each
+   character of output generated by this program so that the beginning
+   and ending indices (in that output) of every file name can be recorded
+   and later output themselves.  */
+static size_t dired_pos;
+
+#define PUTCHAR(c) do {putchar ((c)); ++dired_pos;} while (0)
+
+/* Write S to STREAM and increment DIRED_POS by S_LEN.  */
+#define FPUTS(s, stream, s_len) \
+    do {fputs ((s), (stream)); dired_pos += s_len;} while (0)
+
+/* Like FPUTS, but for use when S is a literal string.  */
+#define FPUTS_LITERAL(s, stream) \
+    do {fputs ((s), (stream)); dired_pos += sizeof((s)) - 1;} while (0)
+
+#define DIRED_INDENT()							\
+    do									\
+      {									\
+	/* FIXME: remove the `&& format == long_format' clause.  */	\
+	if (dired && format == long_format)				\
+	  FPUTS_LITERAL ("  ", stdout);					\
+      }									\
+    while (0)
+
+/* With --dired, store pairs of beginning and ending indices of filenames.  */
+static struct obstack dired_obstack;
+
+/* With --dired and --recursive, store pairs of beginning and ending
+   indices of directory names.  */
+static struct obstack subdired_obstack;
+
+/* Save the current index on the specified obstack, OBS.  */
+#define PUSH_CURRENT_DIRED_POS(obs)					\
+  do									\
+    {									\
+      /* FIXME: remove the `&& format == long_format' clause.  */	\
+      if (dired && format == long_format)				\
+	obstack_grow ((obs), &dired_pos, sizeof (dired_pos));		\
+    }									\
+  while (0)
+
 static enum time_type const time_types[] =
 {
   time_atime, time_atime, time_atime, time_ctime, time_ctime
 };
 
 
+/* Write to standard output the string PREFIX followed by a space-separated
+   list of the integers stored in OS all on one line.  */
+
+static void
+dired_dump_obstack (prefix, os)
+     const char *prefix;
+     struct obstack *os;
+{
+  int i, n_pos;
+  size_t *pos;
+
+  fputs (prefix, stdout);
+  n_pos = obstack_object_size (os) / sizeof (size_t);
+  pos = (size_t *) obstack_finish (os);
+  for (i = 0; i < n_pos; i++)
+    printf (" %d", (int) pos[i]);
+  fputs ("\n", stdout);
+}
+
 void
 main (argc, argv)
      int argc;
@@ -444,22 +538,29 @@ main (argc, argv)
   program_name = argv[0];
   i = decode_switches (argc, argv);
 
-  if (flag_version)
+  if (show_version)
     {
-      fprintf (stderr, "%s\n", version_string);
+      printf ("%s\n", version_string);
       exit (0);
     }
 
-  if (flag_help)
-    usage ();
+  if (show_help)
+    usage (0);
 
   format_needs_stat = sort_type == sort_time || sort_type == sort_size
     || format == long_format
     || trace_links || trace_dirs || indicator_style != none
     || print_block_size || print_inode;
 
+  if (dired && format == long_format)
+    {
+      obstack_init (&dired_obstack);
+      if (trace_dirs)
+	obstack_init (&subdired_obstack);
+    }
+
   nfiles = 100;
-  files = (struct file *) xmalloc (sizeof (struct file) * nfiles);
+  files = (struct fileinfo *) xmalloc (sizeof (struct fileinfo) * nfiles);
   files_index = 0;
 
   clear_files ();
@@ -488,7 +589,7 @@ main (argc, argv)
     {
       print_current_files ();
       if (pending_dirs)
-	putchar ('\n');
+	PUTCHAR ('\n');
     }
   else if (pending_dirs && pending_dirs->next == 0)
     print_dir_name = 0;
@@ -505,6 +606,14 @@ main (argc, argv)
       print_dir_name = 1;
     }
 
+  if (dired && format == long_format)
+    {
+      /* No need to free these since we're about to exit.  */
+      dired_dump_obstack ("//DIRED//", &dired_obstack);
+      if (trace_dirs)
+	dired_dump_obstack ("//SUBDIRED//", &subdired_obstack);
+    }
+
   exit (exit_status);
 }
 
@@ -525,29 +634,37 @@ decode_switches (argc, argv)
 
   /* initialize all switches to default settings */
 
-#ifdef MULTI_COL
-  /* This is for the `dir' program.  */
-  format = many_per_line;
-  quote_funny_chars = 1;
-#else
-#ifdef LONG_FORMAT
-  /* This is for the `vdir' program.  */
-  format = long_format;
-  quote_funny_chars = 1;
-#else
-  /* This is for the `ls' program.  */
-  if (isatty (1))
+  switch (ls_mode)
     {
+    case LS_MULTI_COL:
+      /* This is for the `dir' program.  */
       format = many_per_line;
-      qmark_funny_chars = 1;
+      quote_funny_chars = 1;
+      break;
+
+    case LS_LONG_FORMAT:
+      /* This is for the `vdir' program.  */
+      format = long_format;
+      quote_funny_chars = 1;
+      break;
+
+    case LS_LS:
+      /* This is for the `ls' program.  */
+      if (isatty (1))
+	{
+	  format = many_per_line;
+	  qmark_funny_chars = 1;
+	}
+      else
+	{
+	  format = one_per_line;
+	  qmark_funny_chars = 0;
+	}
+      break;
+
+    default:
+      abort ();
     }
-  else
-    {
-      format = one_per_line;
-      qmark_funny_chars = 0;
-    }
-#endif
-#endif
 
   time_type = time_mtime;
   full_time = 0;
@@ -581,7 +698,7 @@ decode_switches (argc, argv)
   p = getenv ("TABSIZE");
   tabsize = p ? atoi (p) : 8;
 
-  while ((c = getopt_long (argc, argv, "abcdfgiklmnpqrstuw:xABCFGI:LNQRST:UX1",
+  while ((c = getopt_long (argc, argv, "abcdfgiklmnpqrstuw:xABCDFGI:LNQRST:UX1",
 			   long_options, (int *) 0)) != EOF)
     {
       switch (c)
@@ -593,16 +710,16 @@ decode_switches (argc, argv)
 	  all_files = 1;
 	  really_all_files = 1;
 	  break;
-	  
+
 	case 'b':
 	  quote_funny_chars = 1;
 	  qmark_funny_chars = 0;
 	  break;
-	  
+
 	case 'c':
 	  time_type = time_ctime;
 	  break;
-	  
+
 	case 'd':
 	  immediate_dirs = 1;
 	  break;
@@ -625,106 +742,110 @@ decode_switches (argc, argv)
 	case 'i':
 	  print_inode = 1;
 	  break;
-	  
+
 	case 'k':
 	  kilobyte_blocks = 1;
 	  break;
-	  
+
 	case 'l':
 	  format = long_format;
 	  break;
-	  
+
 	case 'm':
 	  format = with_commas;
 	  break;
-	  
+
 	case 'n':
 	  numeric_users = 1;
 	  break;
-	  
+
 	case 'p':
 	  indicator_style = not_programs;
 	  break;
-	  
+
 	case 'q':
 	  qmark_funny_chars = 1;
 	  quote_funny_chars = 0;
 	  break;
-	  
+
 	case 'r':
 	  sort_reverse = 1;
 	  break;
-	  
+
 	case 's':
 	  print_block_size = 1;
 	  break;
-	  
+
 	case 't':
 	  sort_type = sort_time;
 	  break;
-	  
+
 	case 'u':
 	  time_type = time_atime;
 	  break;
-	  
+
 	case 'w':
 	  line_length = atoi (optarg);
 	  if (line_length < 1)
 	    error (1, 0, "invalid line width: %s", optarg);
 	  break;
-	  
+
 	case 'x':
 	  format = horizontal;
 	  break;
-	  
+
 	case 'A':
 	  all_files = 1;
 	  break;
-	  
+
 	case 'B':
 	  add_ignore_pattern ("*~");
 	  add_ignore_pattern (".*~");
 	  break;
-	  
+
 	case 'C':
 	  format = many_per_line;
 	  break;
-	  
+
+	case 'D':
+	  dired = 1;
+	  break;
+
 	case 'F':
 	  indicator_style = all;
 	  break;
-	  
+
 	case 'G':		/* inhibit display of group info */
 	  inhibit_group = 1;
 	  break;
-	  
+
 	case 'I':
 	  add_ignore_pattern (optarg);
 	  break;
-	  
+
 	case 'L':
 	  trace_links = 1;
 	  break;
-	  
+
 	case 'N':
 	  quote_funny_chars = 0;
 	  qmark_funny_chars = 0;
 	  break;
-	  
+
 	case 'Q':
 	  quote_as_string = 1;
 	  quote_funny_chars = 1;
 	  qmark_funny_chars = 0;
 	  break;
-	  
+
 	case 'R':
 	  trace_dirs = 1;
 	  break;
-	  
+
 	case 'S':
 	  sort_type = sort_size;
 	  break;
-	  
+
 	case 'T':
 	  tabsize = atoi (optarg);
 	  if (tabsize < 1)
@@ -748,7 +869,7 @@ decode_switches (argc, argv)
 	  if (i < 0)
 	    {
 	      invalid_arg ("sort type", optarg, i);
-	      usage ();
+	      usage (1);
 	    }
 	  sort_type = sort_types[i];
 	  break;
@@ -758,7 +879,7 @@ decode_switches (argc, argv)
 	  if (i < 0)
 	    {
 	      invalid_arg ("time type", optarg, i);
-	      usage ();
+	      usage (1);
 	    }
 	  time_type = time_types[i];
 	  break;
@@ -768,13 +889,13 @@ decode_switches (argc, argv)
 	  if (i < 0)
 	    {
 	      invalid_arg ("format type", optarg, i);
-	      usage ();
+	      usage (1);
 	    }
 	  format = formats[i];
 	  break;
-	  
+
 	default:
-	  usage ();
+	  usage (1);
 	}
     }
 
@@ -810,8 +931,8 @@ queue_directory (name, realname)
 
 static void
 print_dir (name, realname)
-     char *name;
-     char *realname;
+     const char *name;
+     const char *realname;
 {
   register DIR *reading;
   register struct dirent *next;
@@ -853,20 +974,30 @@ print_dir (name, realname)
 
   if (print_dir_name)
     {
-      if (realname)
-	printf ("%s:\n", realname);
-      else
-	printf ("%s:\n", name);
+      const char *dir;
+
+      DIRED_INDENT ();
+      dir = (realname ? realname : name);
+      PUSH_CURRENT_DIRED_POS (&subdired_obstack);
+      FPUTS (dir, stdout, strlen (dir));
+      PUSH_CURRENT_DIRED_POS (&subdired_obstack);
+      FPUTS_LITERAL (":\n", stdout);
     }
 
   if (format == long_format || print_block_size)
-    printf ("total %u\n", total_blocks);
+    {
+      char buf[6 + 20 + 1 + 1];
+
+      DIRED_INDENT ();
+      sprintf (buf, "total %u\n", total_blocks);
+      FPUTS (buf, stdout, strlen (buf));
+    }
 
   if (files_index)
     print_current_files ();
 
   if (pending_dirs)
-    putchar ('\n');
+    PUTCHAR ('\n');
 }
 
 /* Add `pattern' to the list of patterns for which files that match are
@@ -944,7 +1075,7 @@ gobble_file (name, explicit_arg, dirname)
   if (files_index == nfiles)
     {
       nfiles *= 2;
-      files = (struct file *) xrealloc (files, sizeof (struct file) * nfiles);
+      files = (struct fileinfo *) xrealloc (files, sizeof (*files) * nfiles);
     }
 
   files[files_index].linkname = 0;
@@ -964,14 +1095,14 @@ gobble_file (name, explicit_arg, dirname)
 
       if (trace_links)
 	{
-	  val = stat (path, &files[files_index].stat);
+	  val = SAFE_STAT (path, &files[files_index].stat);
 	  if (val < 0)
 	    /* Perhaps a symbolically-linked to file doesn't exist; stat
 	       the link instead. */
-	    val = lstat (path, &files[files_index].stat);
+	    val = SAFE_LSTAT (path, &files[files_index].stat);
 	}
       else
-	val = lstat (path, &files[files_index].stat);
+	val = SAFE_LSTAT (path, &files[files_index].stat);
       if (val < 0)
 	{
 	  error (0, errno, "%s", path);
@@ -994,7 +1125,7 @@ gobble_file (name, explicit_arg, dirname)
 	  if (linkpath
 	      && ((explicit_arg && format != long_format)
 		   || indicator_style != none)
-	      && stat (linkpath, &linkstats) == 0)
+	      && SAFE_STAT (linkpath, &linkstats) == 0)
 	    {
 	      /* Symbolic links to directories that are mentioned on the
 		 command line are automatically traced if not being
@@ -1064,33 +1195,24 @@ gobble_file (name, explicit_arg, dirname)
 static void
 get_link_name (filename, f)
      char *filename;
-     struct file *f;
+     struct fileinfo *f;
 {
-  register char *linkbuf;
-#ifdef _AIX
-  register int bufsiz = PATH_MAX; /* st_size is wrong. */
-#else
-  register int bufsiz = f->stat.st_size;
-#endif
+  char *linkbuf;
   register int linksize;
 
-  linkbuf = (char *) xmalloc (bufsiz + 1);
+  linkbuf = (char *) alloca (PATH_MAX + 2);
   /* Some automounters give incorrect st_size for mount points.
      I can't think of a good workaround for it, though.  */
-  linksize = readlink (filename, linkbuf, bufsiz);
+  linksize = readlink (filename, linkbuf, PATH_MAX + 1);
   if (linksize < 0)
     {
       error (0, errno, "%s", filename);
       exit_status = 1;
-      free (linkbuf);
     }
   else
     {
-#ifdef _AIX
-      linkbuf = (char *) xrealloc (linkbuf, linksize + 1);
-#endif
       linkbuf[linksize] = '\0';
-      f->linkname = linkbuf;
+      f->linkname = xstrdup (linkbuf);
     }
 }
 
@@ -1218,6 +1340,8 @@ sort_files ()
 	case time_atime:
 	  func = sort_reverse ? rev_cmp_atime : compare_atime;
 	  break;
+	default:
+	  abort ();
 	}
       break;
     case sort_name:
@@ -1229,79 +1353,81 @@ sort_files ()
     case sort_size:
       func = sort_reverse ? rev_cmp_size : compare_size;
       break;
+    default:
+      abort ();
     }
 
-  qsort (files, files_index, sizeof (struct file), func);
+  qsort (files, files_index, sizeof (struct fileinfo), func);
 }
 
 /* Comparison routines for sorting the files. */
 
 static int
 compare_ctime (file1, file2)
-     struct file *file1, *file2;
+     struct fileinfo *file1, *file2;
 {
   return longdiff (file2->stat.st_ctime, file1->stat.st_ctime);
 }
 
 static int
 rev_cmp_ctime (file2, file1)
-     struct file *file1, *file2;
+     struct fileinfo *file1, *file2;
 {
   return longdiff (file2->stat.st_ctime, file1->stat.st_ctime);
 }
 
 static int
 compare_mtime (file1, file2)
-     struct file *file1, *file2;
+     struct fileinfo *file1, *file2;
 {
   return longdiff (file2->stat.st_mtime, file1->stat.st_mtime);
 }
 
 static int
 rev_cmp_mtime (file2, file1)
-     struct file *file1, *file2;
+     struct fileinfo *file1, *file2;
 {
   return longdiff (file2->stat.st_mtime, file1->stat.st_mtime);
 }
 
 static int
 compare_atime (file1, file2)
-     struct file *file1, *file2;
+     struct fileinfo *file1, *file2;
 {
   return longdiff (file2->stat.st_atime, file1->stat.st_atime);
 }
 
 static int
 rev_cmp_atime (file2, file1)
-     struct file *file1, *file2;
+     struct fileinfo *file1, *file2;
 {
   return longdiff (file2->stat.st_atime, file1->stat.st_atime);
 }
 
 static int
 compare_size (file1, file2)
-     struct file *file1, *file2;
+     struct fileinfo *file1, *file2;
 {
   return longdiff (file2->stat.st_size, file1->stat.st_size);
 }
 
 static int
 rev_cmp_size (file2, file1)
-     struct file *file1, *file2;
+     struct fileinfo *file1, *file2;
 {
   return longdiff (file2->stat.st_size, file1->stat.st_size);
 }
 
 static int
 compare_name (file1, file2)
-     struct file *file1, *file2;
+     struct fileinfo *file1, *file2;
 {
   return strcmp (file1->name, file2->name);
 }
 
 static int
 rev_cmp_name (file2, file1)
-     struct file *file1, *file2;
+     struct fileinfo *file1, *file2;
 {
   return strcmp (file1->name, file2->name);
 }
@@ -1311,7 +1437,7 @@ rev_cmp_name (file2, file1)
 
 static int
 compare_extension (file1, file2)
-     struct file *file1, *file2;
+     struct fileinfo *file1, *file2;
 {
   register char *base1, *base2;
   register int cmp;
@@ -1332,7 +1458,7 @@ compare_extension (file1, file2)
 
 static int
 rev_cmp_extension (file2, file1)
-     struct file *file1, *file2;
+     struct fileinfo *file1, *file2;
 {
   register char *base1, *base2;
   register int cmp;
@@ -1384,7 +1510,7 @@ print_current_files ()
       for (i = 0; i < files_index; i++)
 	{
 	  print_long_format (files + i);
-	  putchar ('\n');
+	  PUTCHAR ('\n');
 	}
       break;
     }
@@ -1392,14 +1518,22 @@ print_current_files ()
 
 static void
 print_long_format (f)
-     struct file *f;
+     struct fileinfo *f;
 {
   char modebuf[20];
   char timebuf[40];
+
+  /* 7 fields that may (worst case be 64-bit integral values) require 20 bytes,
+     10 character mode field,
+     24 characters for the time,
+     9 spaces following each of these fields,
+     and 1 trailing NUL byte.  */
+  char bigbuf[7 * 20 + 10 + 24 + 9 + 1];
+  char *p;
   time_t when;
 
   mode_string (f->stat.st_mode, modebuf);
-  modebuf[10] = 0;
+  modebuf[10] = '\0';
 
   switch (time_type)
     {
@@ -1434,45 +1568,63 @@ print_long_format (f)
       timebuf[16] = 0;
     }
 
+  p = bigbuf;
+
   if (print_inode)
-    printf ("%6u ", f->stat.st_ino);
+    {
+      sprintf (p, "%*lu ", INODE_DIGITS, (unsigned long) f->stat.st_ino);
+      p += strlen (p);
+    }
 
   if (print_block_size)
-    printf ("%*u ", block_size_size,
-	    convert_blocks (ST_NBLOCKS (f->stat), kilobyte_blocks));
+    {
+      sprintf (p, "%*u ", block_size_size,
+	       (unsigned) convert_blocks (ST_NBLOCKS (f->stat),
+					  kilobyte_blocks));
+      p += strlen (p);
+    }
 
   /* The space between the mode and the number of links is the POSIX
      "optional alternate access method flag". */
-  printf ("%s %3u ", modebuf, f->stat.st_nlink);
+  sprintf (p, "%s %3u ", modebuf, (unsigned int) f->stat.st_nlink);
+  p += strlen (p);
 
   if (numeric_users)
-    printf ("%-8u ", (unsigned int) f->stat.st_uid);
+    sprintf (p, "%-8u ", (unsigned int) f->stat.st_uid);
   else
-    printf ("%-8.8s ", getuser (f->stat.st_uid));
+    sprintf (p, "%-8.8s ", getuser (f->stat.st_uid));
+  p += strlen (p);
 
   if (!inhibit_group)
     {
       if (numeric_users)
-	printf ("%-8u ", (unsigned int) f->stat.st_gid);
+	sprintf (p, "%-8u ", (unsigned int) f->stat.st_gid);
       else
-	printf ("%-8.8s ", getgroup (f->stat.st_gid));
+	sprintf (p, "%-8.8s ", getgroup (f->stat.st_gid));
+      p += strlen (p);
     }
 
   if (S_ISCHR (f->stat.st_mode) || S_ISBLK (f->stat.st_mode))
-    printf ("%3u, %3u ", major (f->stat.st_rdev),
-	    minor (f->stat.st_rdev));
+    sprintf (p, "%3u, %3u ", (unsigned) major (f->stat.st_rdev),
+	     (unsigned) minor (f->stat.st_rdev));
   else
-    printf ("%8lu ", f->stat.st_size);
+    sprintf (p, "%8lu ", (unsigned long) f->stat.st_size);
+  p += strlen (p);
 
-  printf ("%s ", full_time ? timebuf : timebuf + 4);
+  sprintf (p, "%s ", full_time ? timebuf : timebuf + 4);
+  p += strlen (p);
 
+  DIRED_INDENT ();
+  FPUTS (bigbuf, stdout, p - bigbuf);
+  PUSH_CURRENT_DIRED_POS (&dired_obstack);
   print_name_with_quoting (f->name);
+  PUSH_CURRENT_DIRED_POS (&dired_obstack);
 
   if (f->filetype == symbolic_link)
     {
       if (f->linkname)
 	{
-	  fputs (" -> ", stdout);
+	  FPUTS_LITERAL (" -> ", stdout);
 	  print_name_with_quoting (f->linkname);
 	  if (indicator_style != none)
 	    print_type_indicator (f->linkmode);
@@ -1482,14 +1634,78 @@ print_long_format (f)
     print_type_indicator (f->stat.st_mode);
 }
 
-static void
-print_name_with_quoting (p)
-     register char *p;
+
+/* Set QUOTED_LENGTH to strlen(P) and return NULL if P == quoted(P).
+   Otherwise, return xmalloc'd storage containing the quoted version
+   of P and set QUOTED_LENGTH to the length of the quoted P.  */
+
+static char *
+quote_filename (p, quoted_length)
+     register const char *p;
+     size_t *quoted_length;
 {
   register unsigned char c;
+  const char *p0 = p;
+  char *quoted, *q;
+  int found_quotable;
+
+  if (!quote_as_string && !quote_funny_chars && !qmark_funny_chars)
+    {
+      *quoted_length = strlen (p);
+      return NULL;
+    }
+
+  found_quotable = 0;
+  for (c = *p; c; c = *++p)
+    {
+      if (quote_funny_chars)
+	{
+	  switch (c)
+	    {
+	    case '\\':
+	    case '\n':
+	    case '\b':
+	    case '\r':
+	    case '\t':
+	    case '\f':
+	    case ' ':
+	    case '"':
+	      found_quotable = 1;
+	      break;
+
+	    default:
+	      /* FIXME: why not just use the ISPRINT macro here?  */
+	      if (!(c > 040 && c < 0177))
+		found_quotable = 1;
+	      break;
+	    }
+	}
+      else
+	{
+	  if (!(c >= 040 && c < 0177) && qmark_funny_chars)
+	    found_quotable = 1;
+	}
+      if (found_quotable)
+	break;
+    }
+
+  if (!found_quotable)
+    {
+      *quoted_length = p - p0;
+      return NULL;
+    }
+
+  p = p0;
+  quoted = xmalloc (4 * strlen (p) + 1);
+  q = quoted;
+
+#define SAVECHAR(c) *q++ = (c)
+#define SAVE_2_CHARS(c12) \
+    do { *q++ = ((c12)[0]); \
+	 *q++ = ((c12)[1]); } while (0)
 
   if (quote_as_string)
-    putchar ('"');
+    SAVECHAR ('"');
 
   while ((c = *p++))
     {
@@ -1498,57 +1714,80 @@ print_name_with_quoting (p)
 	  switch (c)
 	    {
 	    case '\\':
-	      printf ("\\\\");
+	      SAVE_2_CHARS ("\\\\");
 	      break;
 
 	    case '\n':
-	      printf ("\\n");
+	      SAVE_2_CHARS ("\\n");
 	      break;
 
 	    case '\b':
-	      printf ("\\b");
+	      SAVE_2_CHARS ("\\b");
 	      break;
 
 	    case '\r':
-	      printf ("\\r");
+	      SAVE_2_CHARS ("\\r");
 	      break;
 
 	    case '\t':
-	      printf ("\\t");
+	      SAVE_2_CHARS ("\\t");
 	      break;
 
 	    case '\f':
-	      printf ("\\f");
+	      SAVE_2_CHARS ("\\f");
 	      break;
 
 	    case ' ':
-	      printf ("\\ ");
+	      SAVE_2_CHARS ("\\ ");
 	      break;
 
 	    case '"':
-	      printf ("\\\"");
+	      SAVE_2_CHARS ("\\\"");
 	      break;
 
 	    default:
 	      if (c > 040 && c < 0177)
-		putchar (c);
+		SAVECHAR (c);
 	      else
-		printf ("\\%03o", (unsigned int) c);
+		{
+		  char buf[5];
+		  sprintf (buf, "\\%03o", (unsigned int) c);
+		  q = stpcpy (q, buf);
+		}
 	    }
 	}
       else
 	{
 	  if (c >= 040 && c < 0177)
-	    putchar (c);
+	    SAVECHAR (c);
 	  else if (!qmark_funny_chars)
-	    putchar (c);
+	    SAVECHAR (c);
 	  else
-	    putchar ('?');
+	    SAVECHAR ('?');
 	}
     }
 
   if (quote_as_string)
-    putchar ('"');
+    SAVECHAR ('"');
+
+  *quoted_length = q - quoted;
+
+  SAVECHAR ('\0');
+
+  return quoted;
+}
+
+static void
+print_name_with_quoting (p)
+     register char *p;
+{
+  char *quoted;
+  size_t quoted_length;
+
+  quoted = quote_filename (p, &quoted_length);
+  FPUTS (quoted != NULL ? quoted : p, stdout, quoted_length);
+  if (quoted)
+    free (quoted);
 }
 
 /* Print the file name of `f' with appropriate quoting.
@@ -1557,14 +1796,15 @@ print_name_with_quoting (p)
 
 static void
 print_file_name_and_frills (f)
-     struct file *f;
+     struct fileinfo *f;
 {
   if (print_inode)
-    printf ("%6u ", f->stat.st_ino);
+    printf ("%*lu ", INODE_DIGITS, (unsigned long) f->stat.st_ino);
 
   if (print_block_size)
     printf ("%*u ", block_size_size,
-	    convert_blocks (ST_NBLOCKS (f->stat), kilobyte_blocks));
+	    (unsigned) convert_blocks (ST_NBLOCKS (f->stat),
+				       kilobyte_blocks));
 
   print_name_with_quoting (f->name);
 
@@ -1577,38 +1817,38 @@ print_type_indicator (mode)
      unsigned int mode;
 {
   if (S_ISDIR (mode))
-    putchar ('/');
+    PUTCHAR ('/');
 
 #ifdef S_ISLNK
   if (S_ISLNK (mode))
-    putchar ('@');
+    PUTCHAR ('@');
 #endif
 
 #ifdef S_ISFIFO
   if (S_ISFIFO (mode))
-    putchar ('|');
+    PUTCHAR ('|');
 #endif
 
 #ifdef S_ISSOCK
   if (S_ISSOCK (mode))
-    putchar ('=');
+    PUTCHAR ('=');
 #endif
 
   if (S_ISREG (mode) && indicator_style == all
-      && (mode & (S_IEXEC | S_IEXEC >> 3 | S_IEXEC >> 6)))
-    putchar ('*');
+      && (mode & (S_IEXEC | S_IXGRP | S_IXOTH)))
+    PUTCHAR ('*');
 }
 
 static int
 length_of_file_name_and_frills (f)
-     struct file *f;
+     struct fileinfo *f;
 {
   register char *p = f->name;
   register char c;
   register int len = 0;
 
   if (print_inode)
-    len += 7;
+    len += INODE_DIGITS + 1;
 
   if (print_block_size)
     len += 1 + block_size_size;
@@ -1858,19 +2098,66 @@ attach (dest, dirname, name)
 }
 
 static void
-usage ()
+usage (status)
+     int status;
 {
-  fprintf (stderr, "\
-Usage: %s [-abcdgiklmnpqrstuxABCFGLNQRSUX1] [-w cols] [-T cols] [-I pattern]\n\
-       [--all] [--escape] [--directory] [--inode] [--kilobytes] [--literal]\n\
-       [--numeric-uid-gid] [--hide-control-chars] [--reverse] [--size]\n\
-       [--width=cols] [--tabsize=cols] [--almost-all] [--ignore-backups]\n",
-	   program_name);
-  fprintf (stderr, "\
-       [--classify] [--file-type] [--ignore=pattern] [--dereference]\n\
-       [--quote-name] [--recursive] [--sort={none,time,size,extension}]\n\
-       [--format={long,verbose,commas,across,vertical,single-column}]\n\
-       [--time={atime,access,use,ctime,status}] [--no-group]\n\
-       [--help] [--version] [path...]\n");
-  exit (1);
+  if (status != 0)
+    fprintf (stderr, "Try `%s --help' for more information.\n",
+	     program_name);
+  else
+    {
+      printf ("Usage: %s [OPTION]... [PATH]...\n", program_name);
+      printf ("\
+\n\
+  -A, --almost-all           do not list implied . and ..\n\
+  -a, --all                  do not hide entries starting with .\n\
+  -B, --ignore-backups       do not list implied entries ending with ~\n\
+  -b, --escape               print octal escapes for nongraphic characters\n\
+  -C                         list entries by columns\n\
+  -c                         sort by change time; with -l: show ctime\n\
+  -D, --dired                generate output well suited to Emacs' dired mode\n\
+  -d, --directory            list directory entries instead of contents\n\
+  -F, --classify             append a character for typing each entry\n\
+  -f                         do not sort, enable -aU, disable -lst\n\
+      --format=WORD          across -x, commas -m, horizontal -x, long -l,\n\
+                               single-column -1, verbose -l, vertical -C\n\
+      --full-time            list both full date and full time\n");
+
+      printf ("\
+  -G, --no-group             inhibit display of group information\n\
+  -g                         (ignored)\n\
+  -I, --ignore=PATTERN       do not list implied entries matching shell PATTERN\n\
+  -i, --inode                print index number of each file\n\
+  -k, --kilobytes            use 1024 blocks, not 512 despite POSIXLY_CORRECT\n\
+  -L, --dereference          list entries pointed to by symbolic links\n\
+  -l                         use a long listing format\n\
+  -m                         fill width with a comma separated list of entries\n\
+  -N, --literal              do not quote entry names\n\
+  -n, --numeric-uid-gid      list numeric UIDs and GIDs instead of names\n\
+  -p                         append a character for typing each entry\n\
+  -Q, --quote-name           enclose entry names in double quotes\n\
+  -q, --hide-control-chars   print ? instead of non graphic characters\n\
+  -R, --recursive            list subdirectories recursively\n\
+  -r, --reverse              reverse order while sorting\n\
+  -S                         sort by file size\n");
+
+      printf ("\
+  -s, --size                 print block size of each file\n\
+      --sort=WORD            ctime -c, extension -X, none -U, size -S,\n\
+                               status -c, time -t\n\
+      --time=WORD            atime -u, access -u, use -u\n\
+  -T, --tabsize=COLS         assume tab stops at each COLS instead of 8\n\
+  -t                         sort by modification time; with -l: show mtime\n\
+  -U                         do not sort; list entries in directory order\n\
+  -u                         sort by last access time; with -l: show atime\n\
+  -w, --width=COLS           assume screen width instead of current value\n\
+  -x                         list entries by lines instead of by columns\n\
+  -X                         sort alphabetically by entry extension\n\
+  -1                         list one file per line\n\
+      --help                 display this help and exit\n\
+      --version              output version information and exit\n\
+\n\
+Sort entries alphabetically if none of -cftuSUX nor --sort.\n");
+    }
+  exit (status);
 }
